from asyncio import CancelledError, Queue, Task, TaskGroup
from collections.abc import Coroutine
from logging import getLogger
from typing import TypeAlias


Callback: TypeAlias = Coroutine[None, None, None]
AioQueue: TypeAlias = Queue[Callback | None]


def create_queue(maxsize: int = 0) -> AioQueue:
    return Queue(maxsize)


async def consume(q: AioQueue) -> None:
    while True:
        cb = await q.get()
        if not cb:
            break
        try:
            await cb
        finally:
            q.task_done()


async def consume_all(q: AioQueue, maxsize: int = 1):
    if q.empty():
        raise ValueError(f"queue must have something to consume")

    if maxsize <= 0:
        raise ValueError(f"invalid consumer size: {maxsize}")

    async with TaskGroup() as tg:
        consumer_list: list[Task[None]] = []
        while maxsize > 0:
            task = tg.create_task(consume(q))
            consumer_list.append(task)
            maxsize -= 1

        await q.join()
        for task in consumer_list:
            task.cancel()
            try:
                await task
            except CancelledError:
                # have to consume the exception
                getLogger(__name__).debug("consumer cancelled")



from importlib.metadata import version

from .aioqueue import AioQueue, create_queue, consume_all
from .queue import AsyncQueue
from .task import Task


__version__ = version(__package__ or __name__)
__all__ = ("AsyncQueue", "Task", "AioQueue", "create_queue", "consume_all")



import functools
import inspect
import itertools
from typing import Any, Awaitable, Callable, Union


RawTask = Callable[[], Any]
MaybeTask = Union["Task", RawTask]


@functools.total_ordering
class Task(object):
    __counter = itertools.count()

    def __init__(self, callable_: RawTask = None) -> None:
        super(Task, self).__init__()

        self.__callable = callable_
        # FIXME atomic because GIL
        self.__id = next(self.__counter)

    def __eq__(self, that: "Task") -> bool:
        return self.equal(that)

    def __lt__(self, that: "Task") -> bool:
        return self.higher_then(that)

    def __call__(self) -> Any:
        if not self.__callable:
            raise NotImplementedError()
        return self.__callable()

    # highest first
    @property
    def priority(self) -> int:
        return 0

    @property
    def id_(self) -> int:
        return self.__id

    def equal(self, that: "Task") -> bool:
        if not isinstance(that, Task):
            return NotImplemented
        return self.id_ == that.id_

    def higher_then(self, that: "Task") -> bool:
        if not isinstance(that, Task):
            return NotImplemented
        if self.priority > that.priority:
            return True
        if self.priority < that.priority:
            return False
        # lower ID was created earlier
        return self.id_ < that.id_


def ensure_task(maybe_task: MaybeTask) -> Task:
    if not isinstance(maybe_task, Task):
        maybe_task = Task(maybe_task)
    return maybe_task


async def regular_call(task: Task) -> Awaitable[Any]:
    rv = task()
    if inspect.isawaitable(rv):
        rv = await rv
    return rv



import asyncio
from contextlib import asynccontextmanager
from logging import getLogger
from os import cpu_count
from typing import Callable

from .task import regular_call, ensure_task, MaybeTask


class AsyncQueue:
    def __init__(self, maximum: int | None = None):
        if maximum is not None:
            self._max = maximum
        else:
            cpu = cpu_count()
            if not cpu:
                self._max = 1
            else:
                self._max = cpu
        self._waiting_idle = asyncio.Condition()
        self._reset()

    async def __aenter__(self):
        return self

    async def __aexit__(self, type_, exc, tb):
        await self.shutdown()

    async def join(self):
        if self.idle:
            return
        async with self._waiting_idle:
            await self._waiting_idle.wait()

    async def shutdown(self):
        # no consumer means it never used
        if not self._consumer_list or self._shutting_down:
            return

        self._shutting_down = True

        # if some consumers are busy, we need to wait for running tasks
        await self.join()

        # cancel all consumers
        for consumer in self._consumer_list:
            consumer.cancel()
        await asyncio.wait(self._consumer_list)

        self._reset()

    def flush(self, filter_: Callable[["Task"], bool] = None):
        q = self._get_internal_queue()
        if filter_ is not None:
            nq = [_ for _ in q if not filter_(_)]
        else:
            nq = []
        getLogger(__name__).debug(f"flush: before {len(q)} after {len(nq)}")
        self._set_internal_queue(nq)

    def post(self, task: MaybeTask):
        if self._shutting_down:
            return

        self._start()

        task = ensure_task(task)
        self._queue.put_nowait(task)

    @property
    def idle(self):
        q = self._get_internal_queue()
        return not q and self._active_count == 0

    def _start(self):
        if self._consumer_list:
            return

        loop = asyncio.get_running_loop()
        self._consumer_list = [
            loop.create_task(self._consume()) for _ in range(self._max)
        ]

    async def _consume(self):
        while True:
            async with self._pop_task() as task, self._active_guard():
                try:
                    await regular_call(task)
                except Exception:
                    getLogger(__name__).exception("uncaught exception")

            async with self._waiting_idle:
                if self.idle:
                    self._waiting_idle.notify_all()

    def _reset(self):
        self._queue = asyncio.PriorityQueue()
        self._active_count = 0
        self._shutting_down = False
        self._consumer_list = None

    def _get_internal_queue(self):
        return self._queue._queue

    def _set_internal_queue(self, nq):
        self._queue._queue = nq

    @asynccontextmanager
    async def _active_guard(self):
        self._active_count += 1
        try:
            yield
        finally:
            self._active_count -= 1

    @asynccontextmanager
    async def _pop_task(self):
        task = await self._queue.get()
        try:
            yield task
        finally:
            self._queue.task_done()
